from typing import Callable
from typing import Dict
from typing import List

from .http_method import HttpMethod
from .routing import Route
from .routing import Router
from .serverless import IS_SERVERLESS_ENVIRONMENT
from .serverless import make_serverless_callback
from .middleware import Middleware


class HttpApplication:
    """
    should support

    schema - for handling json schema
    timeout - for time - outing
    """
    def __init__(self) -> None:
        self.middleware: List[Middleware] = []
        self.methods: Dict[HttpMethod, Router] = {key: Router() for key in HttpMethod}

    def get(self, route: str) -> Callable:
        def _get(handler: Callable) -> Callable:
            self.methods[HttpMethod.GET].append(Route(route), handler)
            if IS_SERVERLESS_ENVIRONMENT:
                return make_serverless_callback(handler, Route(route))
            return handler

        return _get

    def post(self, route: str) -> Callable:
        def _post(handler: Callable) -> Callable:
            self.methods[HttpMethod.POST].append(Route(route), handler)
            if IS_SERVERLESS_ENVIRONMENT:
                return make_serverless_callback(handler, Route(route))
            return handler

        return _post

    def put(self, route: str) -> Callable:
        def _put(handler: Callable) -> Callable:
            self.methods[HttpMethod.PUT].append(Route(route), handler)
            if IS_SERVERLESS_ENVIRONMENT:
                return make_serverless_callback(handler, Route(route))
            return handler

        return _put

    def patch(self, route: str) -> Callable:
        def _patch(handler: Callable) -> Callable:
            self.methods[HttpMethod.PATCH].append(Route(route), handler)
            if IS_SERVERLESS_ENVIRONMENT:
                return make_serverless_callback(handler, Route(route))
            return handler

        return _patch

    def delete(self, route: str) -> Callable:
        def _delete(handler: Callable) -> Callable:
            self.methods[HttpMethod.DELETE].append(Route(route), handler)
            if IS_SERVERLESS_ENVIRONMENT:
                return make_serverless_callback(handler, Route(route))
            return handler

        return _delete

    def head(self, route: str) -> Callable:
        def _head(handler: Callable) -> Callable:
            self.methods[HttpMethod.HEAD].append(Route(route), handler)
            if IS_SERVERLESS_ENVIRONMENT:
                return make_serverless_callback(handler, Route(route))
            return handler

        return _head

    def options(self, route: str) -> Callable:
        def _options(handler: Callable) -> Callable:
            self.methods[HttpMethod.OPTIONS].append(Route(route), handler)
            if IS_SERVERLESS_ENVIRONMENT:
                return make_serverless_callback(handler, Route(route))
            return handler

        return _options

    def any(self, route: str) -> Callable:
        def _any(handler: Callable) -> Callable:
            self.methods[HttpMethod.GET].append(Route(route), handler)
            self.methods[HttpMethod.POST].append(Route(route), handler)
            self.methods[HttpMethod.PUT].append(Route(route), handler)
            self.methods[HttpMethod.PATCH].append(Route(route), handler)
            self.methods[HttpMethod.DELETE].append(Route(route), handler)
            self.methods[HttpMethod.HEAD].append(Route(route), handler)
            self.methods[HttpMethod.OPTIONS].append(Route(route), handler)
            if IS_SERVERLESS_ENVIRONMENT:
                return make_serverless_callback(handler, Route(route))
            return handler

        return _any


http = HttpApplication()


__all__ = ["HttpApplication", "http"]



import json
from collections import UserDict
from io import BytesIO
from json.decoder import JSONDecodeError
from typing import Any
from typing import Optional

from .http_multipart_message_parser import parse_multipart_message
from .http_query_string import parse_qs


class HttpMessage(str):
    pass


class CompositeHttpMessage(UserDict, HttpMessage):
    def get(self, name: str, default: Optional[Any] = None) -> Any:
        if name in self:
            return self[name]

        return default


class FormHttpMessage(CompositeHttpMessage):
    @staticmethod
    def from_bytes(body: BytesIO, encoding: str = "utf8") -> "FormHttpMessage":
        body.seek(0)
        decoded_input = body.read().decode(encoding)
        fields = parse_qs(decoded_input)
        instance = FormHttpMessage()

        for name, value in fields.items():
            instance[name] = value

        return instance


class JsonHttpMessage(CompositeHttpMessage):
    @staticmethod
    def from_bytes(body: BytesIO, encoding: str = "utf8") -> "JsonHttpMessage":
        body.seek(0)
        decoded_input = body.read().decode(encoding)

        try:
            body = json.loads(decoded_input)
        except JSONDecodeError:
            body = {}

        instance = JsonHttpMessage(body)

        return instance


class MultipartHttpMessage(CompositeHttpMessage):
    @staticmethod
    def from_bytes(body: BytesIO, boundary: str, encoding: str = "utf8") -> "MultipartHttpMessage":
        body.seek(0)
        fields = parse_multipart_message(body.read(), boundary, encoding)

        instance = MultipartHttpMessage()

        for name, value in fields.items():
            instance[name] = value

        return instance


__all__ = ["HttpMessage", "CompositeHttpMessage", "FormHttpMessage", "JsonHttpMessage", "MultipartHttpMessage"]



from enum import Enum
from http import HTTPStatus


class HttpStatus(Enum):
    CONTINUE = 100, 'Continue'
    SWITCHING_PROTOCOLS = 101, 'Switching Protocols'
    PROCESSING = 102, 'Processing'
    OK = 200, 'OK'
    CREATED = 201, 'Created'
    ACCEPTED = 202, 'Accepted'
    NON_AUTHORITATIVE_INFORMATION = 203, 'Non-Authoritative Information'
    NO_CONTENT = 204, 'No Content'
    RESET_CONTENT = 205, 'Reset Content'
    PARTIAL_CONTENT = 206, 'Partial Content'
    MULTI_STATUS = 207, 'Multi-Status'
    ALREADY_REPORTED = 208, 'Already Reported'
    IM_USED = 226, 'IM Used'
    MULTIPLE_CHOICES = 300, 'Multiple Choices'
    MOVED_PERMANENTLY = 301, 'Moved Permanently'
    FOUND = 302, 'Found'
    SEE_OTHER = 303, 'See Other'
    NOT_MODIFIED = 304, 'Not Modified'
    USE_PROXY = 305, 'Use Proxy'
    TEMPORARY_REDIRECT = 307, 'Temporary Redirect'
    PERMANENT_REDIRECT = 308, 'Permanent Redirect'
    BAD_REQUEST = 400, 'Bad Request'
    UNAUTHORIZED = 401, 'Unauthorized'
    PAYMENT_REQUIRED = 402, 'Payment Required'
    FORBIDDEN = 403, 'Forbidden'
    NOT_FOUND = 404, 'Not Found'
    METHOD_NOT_ALLOWED = 405, 'Method Not Allowed'
    NOT_ACCEPTABLE = 406, 'Not Acceptable'
    PROXY_AUTHENTICATION_REQUIRED = 407, 'Proxy Authentication Required'
    REQUEST_TIMEOUT = 408, 'Request Timeout'
    CONFLICT = 409, 'Conflict'
    GONE = 410, 'Gone'
    LENGTH_REQUIRED = 411, 'Length Required'
    PRECONDITION_FAILED = 412, 'Precondition Failed'
    REQUEST_ENTITY_TOO_LARGE = 413, 'Request Entity Too Large'
    REQUEST_URI_TOO_LONG = 414, 'Request-URI Too Long'
    UNSUPPORTED_MEDIA_TYPE = 415, 'Unsupported Media Type'
    REQUESTED_RANGE_NOT_SATISFIABLE = 416, 'Requested Range Not Satisfiable'
    EXPECTATION_FAILED = 417, 'Expectation Failed'
    MISDIRECTED_REQUEST = 421, 'Misdirected Request'
    UNPROCESSABLE_ENTITY = 422, 'Unprocessable Entity'
    LOCKED = 423, 'Locked'
    FAILED_DEPENDENCY = 424, 'Failed Dependency'
    UPGRADE_REQUIRED = 426, 'Upgrade Required'
    PRECONDITION_REQUIRED = 428, 'Precondition Required'
    TOO_MANY_REQUESTS = 429, 'Too Many Requests'
    REQUEST_HEADER_FIELDS_TOO_LARGE = 431, 'Request Header Fields Too Large'
    UNAVAILABLE_FOR_LEGAL_REASONS = 451, 'Unavailable For Legal Reasons'
    INTERNAL_SERVER_ERROR = 500, 'Internal Server Error'
    NOT_IMPLEMENTED = 501, 'Not Implemented'
    BAD_GATEWAY = 502, 'Bad Gateway'
    SERVICE_UNAVAILABLE = 503, 'Service Unavailable'
    GATEWAY_TIMEOUT = 504, 'Gateway Timeout'
    HTTP_VERSION_NOT_SUPPORTED = 505, 'HTTP Version Not Supported'
    VARIANT_ALSO_NEGOTIATES = 506, 'Variant Also Negotiates'
    INSUFFICIENT_STORAGE = 507, 'Insufficient Storage'
    LOOP_DETECTED = 508, 'Loop Detected'
    NOT_EXTENDED = 510, 'Not Extended'
    NETWORK_AUTHENTICATION_REQUIRED = 511, 'Network Authentication Required'

    def __str__(self) -> str:
        return "%d %s" % self.value

    def __int__(self) -> int:
        return self.value[0]

    @classmethod
    def from_int(cls, status: int) -> 'HttpStatus':
        value = [item for item in cls if item.value[0] == status]
        if value:
            return value[0]

        raise ValueError(f"Invalid status value `{status}`")


__all__ = ["HttpStatus"]



from io import BytesIO
from typing import Any
from typing import Callable
from typing import Dict
from typing import Union

from . import HttpResponse
from .application import http
from .http_error import HttpError
from .http_headers import HttpHeaders
from .http_method import HttpMethod
from .http_query_string import HttpQueryString
from .http_request import HttpRequest
from .middleware import Middleware
from .middleware import MiddlewarePipeline
from .router_middleware import RouterMiddleware


def create_http_request_from_wsgi(environ: Dict[str, Any]) -> HttpRequest:
    headers = HttpHeaders()
    for key, value in environ.items():
        if not key.startswith("HTTP"):
            continue
        headers.set(key, value)
    headers.set("Content-Type", environ.get("CONTENT_TYPE", "text/plain"))

    return HttpRequest(
        method=HttpMethod(environ.get("REQUEST_METHOD", "GET").upper()),
        path=environ.get("PATH_INFO", "/"),
        body=environ.get("wsgi.input", BytesIO(b"")),
        query_string=HttpQueryString(environ.get("QUERY_STRING", "")),
        headers=headers,
    )


def create_wsgi_handler(*middleware: Union[Callable, Middleware], debug: bool = False) -> Callable[[Dict[str, Any]], Callable]:

    def _handler(environ: Dict[str, Any], start: Callable) -> bytes:
        # Prepare pipeline
        middleware_pipeline = MiddlewarePipeline()
        for item in middleware:
            middleware_pipeline.append(item)

        middleware_pipeline.append(RouterMiddleware.from_http_application(http))

        request = create_http_request_from_wsgi(environ)

        if debug:
            try:
                response = middleware_pipeline(request)
            except HttpError as http_error:
                response = HttpResponse(http_error.http_message, http_error.status_code)
        else:
            # Always send a response
            try:
                response = middleware_pipeline(request)
            except HttpError as http_error:
                response = HttpResponse(http_error.http_message, http_error.status_code)
            except Exception:
                response = HttpResponse("Internal Server Error", 500)

        start(
            str(int(response.status_code)),
            [(key, value) for key, value in response.headers.items()],
        )

        response.body.seek(0)
        return response.body

    return _handler


def serve(*middleware: Union[Callable, Middleware], host: str = "127.0.0.1", port=80, debug: bool = False) -> None:
    import bjoern

    wsgi_handler = create_wsgi_handler(*middleware, debug=debug)

    bjoern.run(wsgi_handler, host, port)
